.error-page preferencespage { margin: 30px; }

.expander { padding: 12px; }
.expander.expanded { border: 0 solid @borders; border-bottom-width: 1px; }
.expander-toolbar {
  border: 0 solid @borders;
  border-top-width: 1px;
  padding: 3px;
}
