/* Slider */

$slider_size: 16px;

.slider {
  // slider trough
  -barlevel-height: 4px;
  -barlevel-background-color: transparentize($fg_color, 0.8); //background of the trough
  -barlevel-border-width: 2px; 
  -barlevel-border-color: transparent; // trough border color
  // fill style
  -barlevel-active-background-color: $selected_bg_color;
  -barlevel-active-border-color: transparent;
  // overfill style (red in this case)
  -barlevel-overdrive-color: $destructive_color;
  -barlevel-overdrive-border-color: transparent; //trough border when red;
  -barlevel-overdrive-separator-width:1px;
  // slider handler
  -slider-handle-radius: $slider_size * 0.5; // half the size of the size
  -slider-handle-border-width: 0;
  -slider-handle-border-color: transparent; // because 0 width
}
